@@ -3,7 +3,7 @@
     sphinx.directives.code
     ~~~~~~~~~~~~~~~~~~~~~~
 
-    :copyright: Copyright 2007-2009 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2010 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
@@ -155,9 +155,9 @@ class LiteralInclude(Directive):
             use = not startafter
             res = []
             for line in lines:
-                if not use and startafter in line:
+                if not use and startafter and startafter in line:
                     use = True
-                elif use and endbefore in line:
+                elif use and endbefore and endbefore in line:
                     use = False
                     break
                 elif use:
